# YML 2.0.34 backend

# written by VB.

import re
import fileinput
import sys, traceback, exceptions, os
from xml.sax.saxutils import escape, quoteattr
from copy import deepcopy
from glob import glob
from pyPEG import code, parse, parseLine
from yml2 import ymlCStyle, comment, _inner

ymlFunc, pointers, pythonFunc = {}, {}, {}
in_ns = ""
operator = []
included = ""
includePath = []

lq = re.compile(r"\|(\>*)(.*)")
sq = re.compile(r"(\d*)\>(.*)")
ts = re.compile(r'(\|\|(?P<inds>\>*)\s*\n(?P<text1>.*?)\n(?P<base>\s*)\|\|)|("""(?P<text2>.*?)""")|(\>\>(?P<text3>.*?)\>\>)', re.S)
tq = re.compile(r"(\]|\<)\s*(.*)")
bq = re.compile(r"\`(.*?)\`", re.S)
bqq = re.compile(r"\s*\`\`(.*)")
all = re.compile(r".*", re.S)

def pointer(name):
    try:
        return str(pointers[name[1:]])
    except:
        if included:
            raise LookupError("in " + included + ": " + name)
        else:
            raise LookupError(name)

def evalPython(expr):
    try:
        return eval(expr, pythonFunc)
    except:
        name, parm, tb = sys.exc_info()
        msg = "in python expression: " + str(parm)
        if name is exceptions.SyntaxError:
            tbl = traceback.format_exception(name, parm, tb)
            msg += "\n" + tbl[-3] + tbl[-2]
        else:
            msg += ": " + expr + "\n"
        if included:
            raise name("in " + included + ": " + msg)
        else:
            raise name(msg)
    
def execPython(script):
    try:
        exec script in pythonFunc
    except:
        name, parm, tb = sys.exc_info()
        msg = "in python script: " + str(parm)
        if name is exceptions.SyntaxError:
            tbl = traceback.format_exception(name, parm, tb)
            msg += "\n" + tbl[-3] + tbl[-2]
        else:
            msg += ": " + expr + "\n"
        if included:
            raise name("in " + included + ": " + msg)
        else:
            raise name(msg)

def textOut(text):
    if not text:
        return ""
    text = text.replace(r'\"', r'\\"')
    text = '"""' + text.replace('"', r'\"') + '"""'
    try:
        textFunc = ymlFunc["text"]
        parms = ['text', ('parm', [text])]
        c, result = textFunc(parms)
        if c:
            result += "</" + textFunc.alias + ">"
        return result
    except:
        return escape(eval(text))

def strRepl(text):
    if not text:
        return ""
    text = text.replace(r'\"', r'\\"')
    text = '"""' + text.replace('"', r'\"') + '"""'
    return escape(eval(text))

def applyMacros(macros, text):
    result = text
    for key, value in macros.iteritems():
        result = result.replace(key, value)
    return result

class YF:
    def __init__(self, name):
        self.parms = []
        self.descends = []
        self.values = {}
        self.content = None
        self.pointers = {}
        self.macros = {}
        if in_ns:
            self.alias = in_ns + ":" + name.replace("_", "-")
        else:
            self.alias = name.replace("_", "-")
        pythonFunc["yml_" + name] = self

    def copy(self, newName):
        yf = YF(newName)
        yf.parms.extend(self.parms)
        yf.descends.extend(self.descends)
        yf.values = self.values.copy()
        yf.content = self.content
        yf.pointers = self.pointers.copy()
        yf.macros = self.macros.copy()
        yf.alias = self.alias
        return yf

    def __call__(self, called_with, hasContent = False, avoidTag = False):
        global pointers
        parms = []
        vals = {}
        if self.pointers:
            pointers.update(self.pointers)
        cw = called_with
        for desc in self.descends:
            try:
                cw = cw[1][1][0][1]
                vals[desc] = cw[0]
            except:
                raise LookupError(called_with[0] + " needs value for +" + desc)
        for data in cw:
            if type(data) == type(("",)):
                if data[0] == "parm":
                    l = data[1]
                    parm = l[0]
                    if parm[0] == "*":
                        parm = pointer(parm)
                    if len(l) == 1:
                        if type(parm) is type((None,)):
                            if parm[0] == "pyExp":
                                val = evalPython(parm[1][0])
                                parms.append(val)
                        else:
                            parms.append(evalPython(str(parm)))
                    else:
                        if type(parm) is type((None,)):
                            if parm[0] == "pyExp":
                                parm = evalPython(parm[1][0])
                        val = l[1]
                        if type(val) is type((None,)):
                            if val[0] == "pyExp":
                                val = evalPython(val[1][0])
                        if val[0] == "*":
                            val = pointer(val)
                        vals[parm] = evalPython(str(val))
                elif data[0][:7] == "content":
                    hasContent = True
        return self.xml(parms, vals, hasContent, avoidTag)

    def addParm(self, parm):
        if parm[0] == "%":
            for i in range(len(self.parms)):
                if self.parms[i][0] != "%":
                    self.parms.insert(i, parm)
                    return
        self.parms.append(parm)

    def addDescend(self, desc):
        self.descends.append(desc[1:])

    def addValue(self, parm, value):
        self.values[parm] = str(evalPython(str(value)))

    def xml(self, callParms, callValues, hasContent, avoidTag = False):
        global pointers
        extraContent = ""
        if self.content:
            hasContent = True
        resultParms = self.values.copy()
        macros = self.macros.copy()
        toDelete = resultParms.keys()
        for key in toDelete:
            if key[0] == "*":
                del resultParms[key]
        for key, value in callValues.iteritems():
            if key[0] == "%":
                macros[key] = value
            else:
                resultParms[key] = value
        i = 0
        for cp in callParms:
            if i < len(self.parms):
                if self.parms[i][0] == "*":
                    cp = str(cp)
                    if "'" in cp:
                        pointers[self.parms[i][1:]] = '"' + cp + '"'
                    else:
                        pointers[self.parms[i][1:]] = "'" + cp + "'"
                elif self.parms[i][0] == "%":
                    macros[self.parms[i]] = str(cp)
                else:
                    resultParms[self.parms[i]] = cp
            else:
                extraContent += str(cp)
                hasContent = True
            i += 1
        result = ""
        for p, v in resultParms.iteritems():
            result += " "+ p + "=" + quoteattr(applyMacros(macros, str(v)))
        if hasContent:
            if avoidTag:
                return True, strRepl(extraContent)
            else:
                return True, "<" + self.alias + result + ">" + strRepl(extraContent)
        else:
            if avoidTag:
                return False, ""
            else:
                return False, "<" + self.alias + result + "/>"

def replaceContent(tree, subtree):
    n = 0
    while n < len(tree):
        obj = tree[n]
        if obj[0] == "func":
            l = obj[1]
            if l[0][:7] == "content":
                d = 1
                if subtree:
                    for el in subtree:
                        tree.insert(n+d, el)
                        d += 1
                del tree[n]
                n += d
            else:
                try:
                    if l[-1][0][:7] == "content":
                        replaceContent(l[-1][1], subtree)
                except: pass
        n += 1
    return tree

def executeCmd(text):
    for (regex, pattern) in operator:
        match = re.search(regex, text)
        while match:
            cmd = pattern
            opt = match.groups()
            for i in range(len(opt)):
                cmd = cmd.replace("%" + str(i+1), opt[i])
            text = text[:match.start()] + "`" + cmd + "`"+ text[match.end():]
            match = re.search(regex, text)

    result = ""
    m = re.search(bq, text)
    while text and m:
        cmd  = m.group(1)
        head = textOut(text[:m.start()])
        text = text[m.end():]
        try:
            r, rest = parseLine(cmd, _inner(), [], True, comment)
            if rest: raise SyntaxError(cmd)
        except SyntaxError:
            if included:
                raise SyntaxError("in " + included + ":syntax error in executing command: " + cmd.strip())
            else:
                raise SyntaxError("syntax error in executing command: " + cmd.strip())
        inner = finish(r)
        result += head + inner
        m = re.search(bq, text)
    result += textOut(text)

    return result

def codegen(obj):
    global in_ns, pointers, included
    ctype = obj[0]

    if type(obj) is type(code("")):
        return obj

    if ctype == "empty":
        return code("")

    if ctype == "in_ns":
        in_ns = obj[1][0]
        subtree = obj[1]
        for sel in subtree:
            codegen(sel)
        in_ns = ""
        return code("")

    elif ctype == "decl":
        name = ""
        for data in obj[1]:
            if type(data) == type(""):
                name = data
                try:
                    yf = ymlFunc[name]
                except:
                    ymlFunc[name] = YF(name)
                    yf = ymlFunc[name]
                    if in_ns:
                        yf.alias = in_ns + ":" + name
            elif type(data) == type(("",)):
                if data[0] == "base":
                    base = data[1][0]
                    try:
                        yf = ymlFunc[name] = ymlFunc[base].copy(name)
                    except KeyError:
                        if included:
                            raise KeyError("in " + included + ": " + base + " as base for " + name)
                        else:
                            raise KeyError(base + " as base for " + name)
                elif data[0] == "descend":
                    yf.addDescend(data[1])
                elif data[0] == "declParm":
                    l = data[1]
                    parmName = l[0]
                    if len(l)==1:
                        yf.addParm(parmName)
                    else:
                        value = l[1]
                        if parmName[0] != "%":
                            yf.addValue(parmName, value)
                        if parmName[0] == "*":
                            yf.pointers[parmName[1:]] = value
                            yf.addParm(parmName)
                        elif parmName[0] == "%":
                            yf.macros[parmName] = str(evalPython(str(value)))
                            yf.addParm(parmName)
                elif data[0] == "alias":
                    if in_ns:
                        yf.alias = in_ns + ":" + data[1][0]
                    else:
                        yf.alias = data[1][0]
                elif data[0] == "content":
                    yf.content = data[1]

        return code("")

    elif ctype == "func":
        avoidTag = False
        name = obj[1][0]
        if name[0] == "&":
            avoidTag = True
            name = name[1:]
        subtree = None
        hasContent = False

        if name[0] == "*":
            name = pointer(name)
            if name[0] == "&":
                avoidTag = True
                name = name[1:]
        try: ymlFunc[name]
        except: ymlFunc[name] = YF(name)

        try:
            if obj[1][-1][0][:7] == "content":
                subtree = obj[1][-1][1]
        except: pass

        if ymlFunc[name].content:
            hasContent = True
            treetemplate = deepcopy(ymlFunc[name].content)
            subtree = replaceContent(treetemplate, subtree)

        if subtree:
            hasContent = True

        hasContent, result = ymlFunc[name](obj[1], hasContent, avoidTag)
        if len(ymlFunc[name].descends):
            for i in range(len(ymlFunc[name].descends)):
                try:
                    subtree = subtree[0][1][1][1]
                except:
                    subtree = None
        if subtree:
            for sel in subtree:
                result += codegen(sel)

        if hasContent and not(avoidTag):
            result += "</" + ymlFunc[name].alias + ">"
        return code(result)

    elif ctype == "lineQuote" or ctype == "quote" or ctype == "textsection":
        m, text, multiline, base, inds = None, "", False, 0, 0

        if ctype == "lineQuote":
            text = obj[1]
            m = lq.match(text)
            if m:
                inds = len(m.group(1))
                text = m.group(2)[1:]
            else: inds = 0
        elif ctype == "quote":
            inds = -1
            text = obj[1][0]
            m = sq.match(text)
            if m:
                if m.group(1):
                    inds = int(m.group(1))
                text = m.group(2)[1:]
            else:
                if type(text) == type(""):
                    text = str(evalPython(text))
        elif ctype == "textsection":
            text = obj[1]
            m = ts.match(text)
            if m.group("text1"):
                inds = 0
                text = m.group("text1")
                multiline = True
                if m.group("base"):
                    base = len(m.group("base"))
                if m.group("inds"):
                    inds = len(m.group("inds"))
            elif m.group("text2"):
                inds = -1
                text = m.group("text2")
            else:
                inds = -1
                text = ""
                if m.group("text3"):
                    for part in m.group("text3").splitlines():
                        if text: text += " "
                        text += part.strip()

        ind = ""
        if inds > -1:
            try:
                cmd = evalPython("indent(" + str(inds) + ")")
                result, rest = parseLine(str(cmd), _inner(), [], True, comment)
                if rest:
                    raise SyntaxError()
                ind = finish(result)
            except: pass
        
        if ctype == "lineQuote": text += "\n"

        hasTextFunc = False
        try:
            ymlFunc["text"]
            hasTextFunc = True
        except: pass

        if multiline:
            result = ""
            indent = None
            try: indent = pythonFunc["indent"]
            except: pass
            cmd = ""

            for line in text.splitlines():
                m = re.match(bqq, line)
                if m:
                    cmd = m.group(1)
                    try:
                        r, x = parseLine(cmd, _inner(), [], True, comment)
                        if x: raise SyntaxError(cmd)
                        result += finish(r)
                    except SyntaxError:
                        if included:
                            raise SyntaxError("in " + included + ": syntax error in executing command: " + cmd.strip())
                        else:
                            raise SyntaxError("syntax error in executing command: " + cmd.strip())
                else:
                    if indent:
                        try:
                            cmd = indent(inds)
                            r, x = parseLine(str(cmd), _inner(), [], True, comment)
                            if x: raise SyntaxError(cmd)
                            result += finish(cmd)
                        except: pass
                    result += executeCmd(line[base:] + "\n")

            text, ind = result, ""

        else:
            text = executeCmd(text)

        return code(ind + text) 

    elif ctype == "tagQuote":
        m = tq.match(obj[1])
        if m.group(1) == "<":
            return code("<" + m.group(2))
        else:
            return code(m.group(2))

    elif ctype == "operator":
        operator.append((re.compile(evalPython(obj[1][0])), obj[1][1]))
        return code("")

    elif ctype == "include":
        reverse = False
        ktext = False
        for arg in obj[1]:
            if type(arg) is type((None,)):
                if arg[0] == "reverse":
                    reverse = True
                elif arg[0] == "ktext":
                    ktext = True
            elif type(arg) is type(""):
                filemask = arg

        if filemask[0] == '/' or filemask[0] == '.':
            files = glob(filemask)
        else:
            files = []
            for dir in includePath:
                path = os.path.join(dir, filemask)
                files.extend(glob(path))

        if reverse:
            files = files[-1::-1]

        if not(files):
            if included:
                raise IOError("in " + included + ": include file(s) '" + filemask + "' not found")
            else:
                raise IOError("include file(s) '" + filemask + "' not found")

        includeFile = fileinput.FileInput(files)
        _included = included
        if ktext:
            text = ""
            for line in includeFile:
                included = includeFile.filename()
                text += executeCmd(line)
            included = _included
            return code(text)
        else:
            included = filemask
            result = parse(ymlCStyle(), includeFile, True, comment)
            x = finish(result)
            included = _included
            return code(x)

    elif ctype == "pyExp":
        exp = obj[1][0]
        cmd = evalPython(exp)
        result, rest = parseLine(str(cmd), _inner(), [], True, comment)
        if rest:
            raise SyntaxError(cmd)
        return code(finish(result))

    elif ctype == "pythonCall":
        parms = []
        data = obj[1]
        for p in data:
            if type(p) is type(""):
                name = p
            elif type(p) is type((None,)):
                ptype = p[0]
                if ptype == "parm":
                    if p[1][0][0] == "*":
                        parms.append(pointer(p[1][0]))
                    else:
                        parms.append(p[1][0])
        if len(parms) == 0:
            exp = name + "()"
        elif len(parms) == 1:
            exp = name + "(" + str(parms[0]) + ")"
        else:
            exp = name + str(tuple(parms))
        cmd = evalPython(exp)
        result, rest = parseLine(str(cmd), _inner(), [], True, comment)
        if rest:
            raise SyntaxError()
        return code(finish(result))

    else:
        return code("")

def finish(tree):
    result = ""
    python = ""

    for el in tree:
        if el[0] == "python":
            if el[1][0][:2] == "!!":
                python += el[1][0][2:-2]
            else:
                python += el[1][0][1:] + "\n"
            continue
        else:
            if python:
                execPython(python)
                python = ""

        result += codegen(el)

    if python:
        execPython(python)

    return result
